from datetime import datetime
from itertools import chain
from math import ceil
from typing import Any, Optional, Iterable, Iterator, List, Dict

# from monty.serialization import loadfn, dumpfn

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.molecules.summary import MoleculeSummaryDoc
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()


class SummaryBuilder(Builder):
    """
    The SummaryBuilder collects all property documents and gathers their properties
    into a single MoleculeSummaryDoc

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. For each doc, grab the relevant property docs
        3. Convert property docs to MoleculeSummaryDoc
    """

    def __init__(
        self,
        molecules: Store,
        charges: Store,
        spins: Store,
        bonds: Store,
        metal_binding: Store,
        orbitals: Store,
        redox: Store,
        thermo: Store,
        vibes: Store,
        summary: Store,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        self.molecules = molecules
        self.charges = charges
        self.spins = spins
        self.bonds = bonds
        self.metal_binding = metal_binding
        self.orbitals = orbitals
        self.redox = redox
        self.thermo = thermo
        self.vibes = vibes
        self.summary = summary
        self.query = query if query else dict()
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(
            sources=[
                molecules,
                charges,
                spins,
                bonds,
                metal_binding,
                orbitals,
                redox,
                thermo,
                vibes,
            ],
            targets=[summary],
            **kwargs,
        )
        # Uncomment in case of issue with mrun not connecting automatically to collections
        # for i in [
        #     self.molecules,
        #     self.charges,
        #     self.spins,
        #     self.bonds,
        #     self.metal_binding,
        #     self.orbitals,
        #     self.redox,
        #     self.thermo,
        #     self.vibes,
        #     self.summary
        # ]:
        #     try:
        #         i.connect()
        #     except Exception as e:
        #         print("Could not connect,", e)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for charges
        self.charges.ensure_index("molecule_id")
        self.charges.ensure_index("method")
        self.charges.ensure_index("task_id")
        self.charges.ensure_index("solvent")
        self.charges.ensure_index("lot_solvent")
        self.charges.ensure_index("property_id")
        self.charges.ensure_index("last_updated")
        self.charges.ensure_index("formula_alphabetical")

        # Search index for charges
        self.spins.ensure_index("molecule_id")
        self.spins.ensure_index("method")
        self.spins.ensure_index("task_id")
        self.spins.ensure_index("solvent")
        self.spins.ensure_index("lot_solvent")
        self.spins.ensure_index("property_id")
        self.spins.ensure_index("last_updated")
        self.spins.ensure_index("formula_alphabetical")

        # Search index for charges
        self.bonds.ensure_index("molecule_id")
        self.bonds.ensure_index("method")
        self.bonds.ensure_index("task_id")
        self.bonds.ensure_index("solvent")
        self.bonds.ensure_index("lot_solvent")
        self.bonds.ensure_index("property_id")
        self.bonds.ensure_index("last_updated")
        self.bonds.ensure_index("formula_alphabetical")

        # Search index for metal_binding
        self.metal_binding.ensure_index("molecule_id")
        self.metal_binding.ensure_index("solvent")
        self.metal_binding.ensure_index("lot_solvent")
        self.metal_binding.ensure_index("property_id")
        self.metal_binding.ensure_index("last_updated")
        self.metal_binding.ensure_index("formula_alphabetical")
        self.metal_binding.ensure_index("method")

        # Search index for orbitals
        self.orbitals.ensure_index("molecule_id")
        self.orbitals.ensure_index("task_id")
        self.orbitals.ensure_index("solvent")
        self.orbitals.ensure_index("lot_solvent")
        self.orbitals.ensure_index("property_id")
        self.orbitals.ensure_index("last_updated")
        self.orbitals.ensure_index("formula_alphabetical")

        # Search index for orbitals
        self.redox.ensure_index("molecule_id")
        self.redox.ensure_index("task_id")
        self.redox.ensure_index("solvent")
        self.redox.ensure_index("lot_solvent")
        self.redox.ensure_index("property_id")
        self.redox.ensure_index("last_updated")
        self.redox.ensure_index("formula_alphabetical")

        # Search index for thermo
        self.thermo.ensure_index("molecule_id")
        self.thermo.ensure_index("task_id")
        self.thermo.ensure_index("solvent")
        self.thermo.ensure_index("lot_solvent")
        self.thermo.ensure_index("property_id")
        self.thermo.ensure_index("last_updated")
        self.thermo.ensure_index("formula_alphabetical")

        # Search index for vibrational properties
        self.vibes.ensure_index("molecule_id")
        self.vibes.ensure_index("task_id")
        self.vibes.ensure_index("solvent")
        self.vibes.ensure_index("lot_solvent")
        self.vibes.ensure_index("property_id")
        self.vibes.ensure_index("last_updated")
        self.vibes.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.summary.ensure_index("molecule_id")
        self.summary.ensure_index("last_updated")
        self.summary.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.summary.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into summary documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the summary Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Summary builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.summary.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into a MoleculeSummaryDoc

        Args:
            tasks List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new orbital docs
        """

        def _group_docs(docs: List[Dict[str, Any]], by_method: bool = False):
            """Helper function to group docs by solvent"""
            grouped: Dict[str, Any] = dict()

            for doc in docs:
                solvent = doc.get("solvent")
                method = doc.get("method")
                if not solvent:
                    # Need to group by solvent
                    continue
                if by_method and method is None:
                    # Trying to group by method, but no method present
                    continue

                if not by_method:
                    grouped[solvent] = doc
                else:
                    if solvent not in grouped:
                        grouped[solvent] = {method: doc}
                    else:
                        grouped[solvent][method] = doc

            return (grouped, by_method)

        mols = items
        formula = mols[0]["formula_alphabetical"]
        mol_ids = [m["molecule_id"] for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        summary_docs = list()

        for mol in mols:
            mol_id = mol["molecule_id"]

            d = {
                "molecules": mol,
                "partial_charges": _group_docs(
                    list(self.charges.query({"molecule_id": mol_id})), True
                ),
                "partial_spins": _group_docs(
                    list(self.spins.query({"molecule_id": mol_id})), True
                ),
                "bonding": _group_docs(
                    list(self.bonds.query({"molecule_id": mol_id})), True
                ),
                "metal_binding": _group_docs(
                    list(self.metal_binding.query({"molecule_id": mol_id})), True
                ),
                "orbitals": _group_docs(
                    list(self.orbitals.query({"molecule_id": mol_id})), False
                ),
                "redox": _group_docs(
                    list(self.redox.query({"molecule_id": mol_id})), False
                ),
                "thermo": _group_docs(
                    list(self.thermo.query({"molecule_id": mol_id})), False
                ),
                "vibration": _group_docs(
                    list(self.vibes.query({"molecule_id": mol_id})), False
                ),
            }

            to_delete = list()

            for k, v in d.items():
                if isinstance(v, dict) and len(v) == 0:
                    to_delete.append(k)

            for td in to_delete:
                del d[td]

            # # For debugging; keep because it might be needed again
            # dumpfn(d, f"{mol_id}.json.gz")
            # break

            summary_doc = MoleculeSummaryDoc.from_docs(molecule_id=mol_id, docs=d)
            summary_docs.append(summary_doc)

        self.logger.debug(f"Produced {len(summary_docs)} summary docs for {formula}")

        return jsanitize([doc.dict() for doc in summary_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new documents into the summary collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} summary documents")
            self.summary.remove_docs({self.summary.key: {"$in": molecule_ids}})
            self.summary.update(
                docs=docs,
                key=["molecule_id"],
            )
        else:
            self.logger.info("No items to update")



import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.qchem.molecules import MoleculesAssociationBuilder, MoleculesBuilder
from emmet.builders.molecules.atomic import PartialChargesBuilder, PartialSpinsBuilder


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "C2H4.json.gz")


@pytest.fixture(scope="session")
def mol_store(tasks_store):
    assoc_store = MemoryStore(key="molecule_id")
    stage_one = MoleculesAssociationBuilder(tasks=tasks_store, assoc=assoc_store)
    stage_one.run()

    mol_store = MemoryStore(key="molecule_id")
    stage_two = MoleculesBuilder(assoc=assoc_store, molecules=mol_store)
    stage_two.run()

    return mol_store


@pytest.fixture(scope="session")
def charges_store():
    return MemoryStore()


@pytest.fixture(scope="session")
def spins_store():
    return MemoryStore()


def test_charges_builder(tasks_store, mol_store, charges_store):
    builder = PartialChargesBuilder(
        tasks_store,
        mol_store,
        charges_store,
        methods=["mulliken", "resp", "critic2", "nbo"],
    )
    builder.run()

    assert charges_store.count() == 80
    assert charges_store.count({"deprecated": True}) == 0


def test_spins_builder(tasks_store, mol_store, spins_store):
    builder = PartialSpinsBuilder(
        tasks_store, mol_store, spins_store, methods=["mulliken", "nbo"]
    )
    builder.run()

    assert spins_store.count() == 18
    assert spins_store.count({"deprecated": True}) == 0



from emmet.builders.utils import (
    chemsys_permutations,
    maximal_spanning_non_intersecting_subsets,
    get_hop_cutoff,
)
from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph
from numpy.testing import assert_almost_equal
from monty.serialization import loadfn


def test_maximal_spanning_non_intersecting_subsets():
    assert maximal_spanning_non_intersecting_subsets([{"A"}, {"A", "B"}]) == {
        frozenset(d) for d in [{"A"}, {"B"}]
    }

    assert maximal_spanning_non_intersecting_subsets([{"A", "B"}, {"A", "B", "C"}]) == {
        frozenset(d) for d in [{"A", "B"}, {"C"}]
    }

    assert maximal_spanning_non_intersecting_subsets(
        [{"A", "B"}, {"A", "B", "C"}, {"D"}]
    ) == {frozenset(d) for d in [{"A", "B"}, {"C"}, {"D"}]}


def test_chemsys_permutations(test_dir):
    assert len(chemsys_permutations("Sr")) == 1
    assert len(chemsys_permutations("Sr-Hf")) == 3
    assert len(chemsys_permutations("Sr-Hf-O")) == 7


def test_get_hop_cutoff(test_dir):
    spinel_mg = loadfn(test_dir / "mobility/migration_graph_spinel_MgMn2O4.json")
    nasicon_mg = loadfn(test_dir / "mobility/migration_graph_nasicon_MgV2(PO4)3.json")

    # tests for "min_distance" algorithm
    assert_almost_equal(
        get_hop_cutoff(spinel_mg.structure, "Mg", algorithm="min_distance"),
        1.95,
        decimal=2,
    )
    assert_almost_equal(
        get_hop_cutoff(nasicon_mg.structure, "Mg", algorithm="min_distance"),
        3.80,
        decimal=2,
    )

    # test for "hops_based" algorithm, terminated by number of unique hops condition
    d = get_hop_cutoff(spinel_mg.structure, "Mg", algorithm="hops_based")
    check_mg = MigrationGraph.with_distance(spinel_mg.structure, "Mg", d)
    assert_almost_equal(d, 4.18, decimal=2)
    assert len(check_mg.unique_hops) == 5

    # test for "hops_based" algorithm, terminated by the largest hop length condition
    d = get_hop_cutoff(nasicon_mg.structure, "Mg", algorithm="hops_based")
    check_mg = MigrationGraph.with_distance(nasicon_mg.structure, "Mg", d)
    assert_almost_equal(d, 4.59, decimal=2)
    assert len(check_mg.unique_hops) == 6



import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.qchem.molecules import MoleculesAssociationBuilder, MoleculesBuilder
from emmet.builders.molecules.atomic import PartialChargesBuilder, PartialSpinsBuilder
from emmet.builders.molecules.bonds import BondingBuilder
from emmet.builders.molecules.thermo import ThermoBuilder
from emmet.builders.molecules.metal_binding import MetalBindingBuilder


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


@pytest.fixture(scope="session")
def tasks(test_dir):
    return JSONStore(test_dir / "lithium_carbonate_tasks.json.gz")


@pytest.fixture(scope="session")
def mols(tasks):
    assoc_store = MemoryStore(key="molecule_id")
    stage_one = MoleculesAssociationBuilder(tasks=tasks, assoc=assoc_store)
    stage_one.run()

    mol_store = MemoryStore(key="molecule_id")
    stage_two = MoleculesBuilder(assoc=assoc_store, molecules=mol_store)
    stage_two.run()

    return mol_store


@pytest.fixture(scope="session")
def charges():
    return MemoryStore()


@pytest.fixture(scope="session")
def spins():
    return MemoryStore()


@pytest.fixture(scope="session")
def bonds():
    return MemoryStore()


@pytest.fixture(scope="session")
def thermo():
    return MemoryStore()


@pytest.fixture(scope="session")
def metal_binding():
    return MemoryStore()


def test_charges_builder(tasks, mols, charges, spins, bonds, metal_binding, thermo):
    charge_build = PartialChargesBuilder(tasks, mols, charges)
    charge_build.run()

    spins_build = PartialSpinsBuilder(tasks, mols, spins)
    spins_build.run()

    bonds_build = BondingBuilder(tasks, mols, bonds)
    bonds_build.run()

    thermo_build = ThermoBuilder(tasks, mols, thermo)
    thermo_build.run()

    metal_binding_build = MetalBindingBuilder(
        mols, charges, spins, bonds, thermo, metal_binding
    )
    metal_binding_build.run()

    assert metal_binding.count() == 6



from itertools import combinations, chain
from typing import Tuple, List, Dict, Union

from tqdm import tqdm
from maggma.builders import Builder
from pymatgen.core.structure import Structure
from matminer.datasets import load_dataset
from emmet.core.thermo import ThermoType

from pymatgen.analysis.alloys.core import (
    AlloyPair,
    InvalidAlloy,
    KNOWN_ANON_FORMULAS,
    AlloyMember,
    AlloySystem,
)

# rough sort of ANON_FORMULAS by "complexity"
ANON_FORMULAS = sorted(KNOWN_ANON_FORMULAS, key=lambda af: len(af))

# Combinatorially, cannot StructureMatch every single possible pair of materials
# Use a loose spacegroup for a pre-screen (in addition to standard spacegroup)
LOOSE_SPACEGROUP_SYMPREC = 0.5

# A source of effective masses, should be replaced with MP-provided effective masses.
BOLTZTRAP_DF = load_dataset("boltztrap_mp")


class AlloyPairBuilder(Builder):
    """
    This builder iterates over anonymous_formula and builds AlloyPair.
    It does not look for members of an AlloyPair.
    """

    def __init__(
        self,
        materials,
        thermo,
        electronic_structure,
        provenance,
        oxi_states,
        alloy_pairs,
        thermo_type: Union[ThermoType, str] = ThermoType.GGA_GGA_U_R2SCAN,
    ):
        self.materials = materials
        self.thermo = thermo
        self.electronic_structure = electronic_structure
        self.provenance = provenance
        self.oxi_states = oxi_states
        self.alloy_pairs = alloy_pairs

        t_type = thermo_type if isinstance(thermo_type, str) else thermo_type.value
        valid_types = {*map(str, ThermoType.__members__.values())}
        if invalid_types := {t_type} - valid_types:
            raise ValueError(
                f"Invalid thermo type(s) passed: {invalid_types}, valid types are: {valid_types}"
            )

        self.thermo_type = t_type

        super().__init__(
            sources=[materials, thermo, electronic_structure, provenance, oxi_states],
            targets=[alloy_pairs],
            chunk_size=8,
        )

    def ensure_indexes(self):
        self.alloy_pairs.ensure_index("pair_id")
        self.alloy_pairs.ensure_index("_search.id")
        self.alloy_pairs.ensure_index("_search.formula")
        self.alloy_pairs.ensure_index("_search.member_ids")
        self.alloy_pairs.ensure_index("alloy_pair.chemsys")

    def get_items(self):
        self.ensure_indexes()

        for idx, af in enumerate(ANON_FORMULAS):
            # if af != "AB":
            #     continue

            thermo_docs = self.thermo.query(
                criteria={
                    "formula_anonymous": af,
                    "deprecated": False,
                    "thermo_type": self.thermo_type,
                },
                properties=[
                    "material_id",
                    "energy_above_hull",
                    "formation_energy_per_atom",
                ],
            )

            thermo_docs = {d["material_id"]: d for d in thermo_docs}

            mpids = list(thermo_docs.keys())

            docs = self.materials.query(
                criteria={
                    "material_id": {"$in": mpids},
                    "deprecated": False,
                },  # , "material_id": {"$in": ["mp-804", "mp-661"]}},
                properties=["structure", "material_id", "symmetry.number"],
            )
            docs = {d["material_id"]: d for d in docs}

            electronic_structure_docs = self.electronic_structure.query(
                {"material_id": {"$in": mpids}},
                properties=["material_id", "band_gap", "is_gap_direct"],
            )
            electronic_structure_docs = {
                d["material_id"]: d for d in electronic_structure_docs
            }

            provenance_docs = self.provenance.query(
                {"material_id": {"$in": mpids}},
                properties=["material_id", "theoretical", "database_IDs"],
            )
            provenance_docs = {d["material_id"]: d for d in provenance_docs}

            oxi_states_docs = self.oxi_states.query(
                {"material_id": {"$in": mpids}, "state": "successful"},
                properties=["material_id", "structure"],
            )
            oxi_states_docs = {d["material_id"]: d for d in oxi_states_docs}

            for material_id in mpids:
                d = docs[material_id]

                d["structure"] = Structure.from_dict(d["structure"])

                if material_id in oxi_states_docs:
                    d["structure_oxi"] = Structure.from_dict(
                        oxi_states_docs[material_id]["structure"]
                    )
                else:
                    d["structure_oxi"] = d["structure"]

                # calculate loose space group
                d["spacegroup_loose"] = d["structure"].get_space_group_info(
                    LOOSE_SPACEGROUP_SYMPREC
                )[1]

                d["properties"] = {}
                # patch in BoltzTraP data if present
                row = BOLTZTRAP_DF.loc[BOLTZTRAP_DF["mpid"] == material_id]
                if len(row) == 1:
                    d["properties"]["m_n"] = float(row.m_n)
                    d["properties"]["m_p"] = float(row.m_p)

                if material_id in electronic_structure_docs:
                    for key in ("band_gap", "is_gap_direct"):
                        d["properties"][key] = electronic_structure_docs[material_id][
                            key
                        ]

                for key in ("energy_above_hull", "formation_energy_per_atom"):
                    d["properties"][key] = thermo_docs[material_id][key]

                if material_id in provenance_docs:
                    for key in ("theoretical",):
                        d["properties"][key] = provenance_docs[material_id][key]

            print(
                f"Starting {af} with {len(docs)} materials, anonymous formula {idx} of {len(ANON_FORMULAS)}"
            )

            yield docs

    def process_item(self, item):
        pairs = []
        for mpids in tqdm(list(combinations(item.keys(), 2))):
            if (
                item[mpids[0]]["symmetry"]["number"]
                == item[mpids[1]]["symmetry"]["number"]
            ) or (
                item[mpids[0]]["spacegroup_loose"] == item[mpids[1]]["spacegroup_loose"]
            ):
                # optionally, could restrict based on band gap too (e.g. at least one end-point semiconducting)
                # if (item[mpids[0]]["band_gap"] > 0) or (item[mpids[1]]["band_gap"] > 0):
                try:
                    pair = AlloyPair.from_structures(
                        structures=[
                            item[mpids[0]]["structure"],
                            item[mpids[1]]["structure"],
                        ],
                        structures_with_oxidation_states=[
                            item[mpids[0]]["structure_oxi"],
                            item[mpids[1]]["structure_oxi"],
                        ],
                        ids=[mpids[0], mpids[1]],
                        properties=[
                            item[mpids[0]]["properties"],
                            item[mpids[1]]["properties"],
                        ],
                    )
                    pairs.append(
                        {
                            "alloy_pair": pair.as_dict(),
                            "_search": pair.search_dict(),
                            "pair_id": pair.pair_id,
                        }
                    )
                except InvalidAlloy:
                    pass
                except Exception as exc:
                    print(exc)

        if pairs:
            print(f"Found {len(pairs)} alloy(s)")

        return pairs

    def update_targets(self, items):
        docs = list(chain.from_iterable(items))
        if docs:
            self.alloy_pairs.update(docs)


class AlloyPairMemberBuilder(Builder):
    """
    This builder iterates over available AlloyPairs by chemical system
    and searches for possible members of those AlloyPairs.
    """

    def __init__(self, alloy_pairs, materials, snls, alloy_pair_members):
        self.alloy_pairs = alloy_pairs
        self.materials = materials
        self.snls = snls
        self.alloy_pair_members = alloy_pair_members

        super().__init__(
            sources=[alloy_pairs, materials, snls], targets=[alloy_pair_members]
        )

    def ensure_indexes(self):
        self.alloy_pairs.ensure_index("pair_id")
        self.alloy_pairs.ensure_index("_search.id")
        self.alloy_pairs.ensure_index("_search.formula")
        self.alloy_pairs.ensure_index("_search.member_ids")
        self.alloy_pairs.ensure_index("alloy_pair.chemsys")
        self.alloy_pairs.ensure_index("alloy_pair.anonymous_formula")

    def get_items(self):
        all_alloy_chemsys = set(self.alloy_pairs.distinct("alloy_pair.chemsys"))
        all_known_chemsys = set(self.materials.distinct("chemsys")) | set(
            self.snls.distinct("chemsys")
        )
        possible_chemsys = all_known_chemsys.intersection(all_alloy_chemsys)

        print(
            f"There are {len(all_alloy_chemsys)} alloy chemical systems of which "
            f"{len(possible_chemsys)} may have members."
        )

        for idx, chemsys in enumerate(possible_chemsys):
            pairs = self.alloy_pairs.query(criteria={"alloy_pair.chemsys": chemsys})
            pairs = [AlloyPair.from_dict(d["alloy_pair"]) for d in pairs]

            mp_docs = self.materials.query(
                criteria={"chemsys": chemsys, "deprecated": False},
                properties=["structure", "material_id"],
            )
            mp_structures = {
                d["material_id"]: Structure.from_dict(d["structure"]) for d in mp_docs
            }

            snl_docs = self.snls.query({"chemsys": chemsys})
            snl_structures = {d["snl_id"]: Structure.from_dict(d) for d in snl_docs}

            structures = mp_structures
            structures.update(snl_structures)

            if structures:
                yield (pairs, structures)

    def process_item(self, item: Tuple[List[AlloyPair], Dict[str, Structure]]):
        pairs, structures = item

        all_pair_members = []
        for pair in pairs:
            pair_members = {"pair_id": pair.pair_id, "members": []}
            for db_id, structure in structures.items():
                try:
                    if pair.is_member(structure):
                        db, _ = db_id.split("-")
                        member = AlloyMember(
                            id_=db_id,
                            db=db,
                            composition=structure.composition,
                            is_ordered=structure.is_ordered,
                            x=pair.get_x(structure.composition),
                        )
                        pair_members["members"].append(member.as_dict())
                except Exception as exc:
                    print(f"Exception for {db_id}: {exc}")
            if pair_members["members"]:
                all_pair_members.append(pair_members)

        return all_pair_members

    def update_targets(self, items):
        docs = list(chain.from_iterable(items))
        if docs:
            self.alloy_pair_members.update(docs)


class AlloySystemBuilder(Builder):
    """
    This builder stitches together the results of
    AlloyPairBuilder and AlloyPairMemberBuilder. The output
    of this collection is the one served by the AlloyPair API.
    It also builds AlloySystem.
    """

    def __init__(
        self, alloy_pairs, alloy_pair_members, alloy_pairs_merged, alloy_systems
    ):
        self.alloy_pairs = alloy_pairs
        self.alloy_pair_members = alloy_pair_members
        self.alloy_pairs_merged = alloy_pairs_merged
        self.alloy_systems = alloy_systems

        super().__init__(
            sources=[alloy_pairs, alloy_pair_members],
            targets=[alloy_pairs_merged, alloy_systems],
            chunk_size=8,
        )

    def get_items(self):
        for idx, af in enumerate(ANON_FORMULAS):
            # comment out to only calculate a single anonymous formula for debugging
            # if af != "AB":
            #     continue

            docs = list(self.alloy_pairs.query({"alloy_pair.anonymous_formula": af}))
            pair_ids = [d["pair_id"] for d in docs]
            members = {
                d["pair_id"]: d
                for d in self.alloy_pair_members.query({"pair_id": {"$in": pair_ids}})
            }

            if docs:
                yield docs, members

    def process_item(self, item):
        pair_docs, members = item

        for doc in pair_docs:
            if doc["pair_id"] in members:
                doc["alloy_pair"]["members"] = members[doc["pair_id"]]["members"]
                doc["_search"]["member_ids"] = [
                    m["id_"] for m in members[doc["pair_id"]]["members"]
                ]
            else:
                doc["alloy_pair"]["members"] = []
                doc["_search"]["member_ids"] = []

        pairs = [AlloyPair.from_dict(d["alloy_pair"]) for d in pair_docs]
        systems = AlloySystem.systems_from_pairs(pairs)

        system_docs = [
            {
                "alloy_system": system.as_dict(),
                "alloy_id": system.alloy_id,
                "_search": {"member_ids": [m.id_ for m in system.members]},
            }
            for system in systems
        ]

        for system_doc in system_docs:
            # Too big to store, will need to reconstruct separately from pair_ids
            system_doc["alloy_system"]["alloy_pairs"] = None

        return pair_docs, system_docs

    def update_targets(self, items):
        pair_docs, system_docs = [p for p, s in items], [s for p, s in items]

        pair_docs = list(chain.from_iterable(pair_docs))
        if pair_docs:
            self.alloy_pairs_merged._collection.insert_many(pair_docs)

        system_docs = list(chain.from_iterable(system_docs))
        if system_docs:
            self.alloy_systems._collection.insert_many(system_docs)



from typing import Dict, Optional
from collections import defaultdict

from maggma.builders import MapBuilder
from maggma.core import Store

from emmet.builders.settings import EmmetBuildSettings
from emmet.core.vasp.task_valid import TaskDocument
from emmet.core.vasp.calc_types.enums import CalcType
from emmet.core.vasp.validation import DeprecationMessage, ValidationDoc


class TaskValidator(MapBuilder):
    def __init__(
        self,
        tasks: Store,
        task_validation: Store,
        potcar_hashes: Optional[Dict[CalcType, Dict[str, str]]] = None,
        settings: Optional[EmmetBuildSettings] = None,
        query: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Creates task_types from tasks and type definitions

        Args:
            tasks: Store of task documents
            task_validation: Store of task_types for tasks
            potcar_hashes: Optional dictionary of potcar hash data.
                Mapping is calculation type -> potcar symbol -> hash value.
        """
        self.tasks = tasks
        self.task_validation = task_validation
        self.settings = EmmetBuildSettings.autoload(settings)
        self.query = query
        self.kwargs = kwargs
        self.potcar_hashes = potcar_hashes

        # Set up potcar cache if appropriate
        if self.settings.VASP_VALIDATE_POTCAR_HASHES:
            if not self.potcar_hashes:
                from pymatgen.io.vasp.inputs import PotcarSingle

                hashes = defaultdict(dict)  # type: dict

                for (
                    calc_type,
                    input_set,
                ) in self.settings.VASP_DEFAULT_INPUT_SETS.items():
                    functional = input_set.CONFIG["POTCAR_FUNCTIONAL"]
                    for potcar_symbol in input_set.CONFIG["POTCAR"].values():
                        potcar = PotcarSingle.from_symbol_and_functional(
                            symbol=potcar_symbol, functional=functional
                        )
                        hashes[calc_type][potcar_symbol] = potcar.get_potcar_hash()

                self.potcar_hashes = potcar_hashes
        else:
            self.potcar_hashes = None

        super().__init__(
            source=tasks,
            target=task_validation,
            projection=[
                "orig_inputs",
                "input.hubbards",
                "output.structure",
                "output.bandgap",
                "chemsys",
                "calcs_reversed",
            ],
            query=query,
            **kwargs,
        )

    def unary_function(self, item):
        """
        Find the task_type for the item

        Args:
            item (dict): a (projection of a) task doc
        """
        task_doc = TaskDocument(**item)
        validation_doc = ValidationDoc.from_task_doc(
            task_doc=task_doc,
            kpts_tolerance=self.settings.VASP_KPTS_TOLERANCE,
            kspacing_tolerance=self.settings.VASP_KSPACING_TOLERANCE,
            input_sets=self.settings.VASP_DEFAULT_INPUT_SETS,
            LDAU_fields=self.settings.VASP_CHECKED_LDAU_FIELDS,
            max_allowed_scf_gradient=self.settings.VASP_MAX_SCF_GRADIENT,
            potcar_hashes=self.potcar_hashes,
        )

        bad_tags = list(set(task_doc.tags).intersection(self.settings.DEPRECATED_TAGS))
        if len(bad_tags) > 0:
            validation_doc.warnings.append(f"Manual Deprecation by tags: {bad_tags}")
            validation_doc.valid = False
            validation_doc.reasons.append(DeprecationMessage.MANUAL)

        return validation_doc



import pytest

from maggma.stores import JSONStore, MemoryStore

from emmet.builders.qchem.molecules import MoleculesAssociationBuilder, MoleculesBuilder
from emmet.builders.molecules.atomic import PartialChargesBuilder, PartialSpinsBuilder
from emmet.builders.molecules.bonds import BondingBuilder
from emmet.builders.molecules.metal_binding import MetalBindingBuilder
from emmet.builders.molecules.orbitals import OrbitalBuilder
from emmet.builders.molecules.redox import RedoxBuilder
from emmet.builders.molecules.thermo import ThermoBuilder
from emmet.builders.molecules.vibration import VibrationBuilder
from emmet.builders.molecules.summary import SummaryBuilder


@pytest.fixture(scope="session")
def tasks(test_dir):
    return JSONStore(test_dir / "lithium_carbonate_tasks.json.gz")


@pytest.fixture(scope="session")
def mols(tasks):
    assoc_store = MemoryStore(key="molecule_id")
    stage_one = MoleculesAssociationBuilder(tasks=tasks, assoc=assoc_store)
    stage_one.run()

    mol_store = MemoryStore(key="molecule_id")
    stage_two = MoleculesBuilder(assoc=assoc_store, molecules=mol_store)
    stage_two.run()

    return mol_store


@pytest.fixture(scope="session")
def charges(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def spins(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def bonds(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def metal_binding(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def orbitals(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def redox(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def thermo(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def vibes(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def summary():
    return MemoryStore(key="molecule_id")


def test_summary_doc(
    tasks,
    mols,
    charges,
    spins,
    bonds,
    metal_binding,
    orbitals,
    redox,
    thermo,
    vibes,
    summary,
):
    charge_build = PartialChargesBuilder(tasks, mols, charges)
    charge_build.run()

    spins_build = PartialSpinsBuilder(tasks, mols, spins)
    spins_build.run()

    bonds_build = BondingBuilder(tasks, mols, bonds)
    bonds_build.run()

    orb_build = OrbitalBuilder(tasks, mols, orbitals)
    orb_build.run()

    thermo_build = ThermoBuilder(tasks, mols, thermo)
    thermo_build.run()

    redox_build = RedoxBuilder(tasks, mols, thermo, redox)
    redox_build.run()

    metal_binding_build = MetalBindingBuilder(
        mols, charges, spins, bonds, thermo, metal_binding
    )
    metal_binding_build.run()

    vibe_build = VibrationBuilder(tasks, mols, vibes)
    vibe_build.run()

    builder = SummaryBuilder(
        molecules=mols,
        charges=charges,
        spins=spins,
        bonds=bonds,
        metal_binding=metal_binding,
        orbitals=orbitals,
        redox=redox,
        thermo=thermo,
        vibes=vibes,
        summary=summary,
    )
    builder.run()

    assert summary.count() == 25



from collections import defaultdict
from datetime import datetime
from itertools import chain
from math import ceil
from typing import Optional, Iterable, Iterator, List, Dict

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.qchem.task import TaskDocument
from emmet.core.qchem.molecule import MoleculeDoc, evaluate_lot
from emmet.core.molecules.bonds import MoleculeBondingDoc, BOND_METHODS
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()


class BondingBuilder(Builder):
    """
    The BondingBuilder defines the bonds in a MoleculeDoc.

    Various methods can be used to define bonding, including:
        - OpenBabelNN + metal_edge_extender: Combining the bond detection algorithms in OpenBabel (OpenBabelNN in
            pymatgen) with a heuristic to add metal coordinate bonds (metal_edge_extender
            in pymatgen)
        - critic2: Using critical points of the electron density to define bonds
        - nbo: Using Natural Bonding Orbital analysis to define bonds and other
            interatomic interactions

    NOTE: Only NBO7 can be used to generate bonding. Bonding (especially when metals
        are involved) is unreliable with earlier version of NBO!

    This builder will attempt to build documents for each molecule, in each solvent,
    with each method. For each molecule-solvent-method combination, the highest-quality
    data available (based on level of theory and electronic energy) will be used.

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. For each molecule, group all tasks by solvent.
        3. For each solvent, sort tasks by level of theory and electronic energy
        4. For each method:
            4.1. Find task docs with necessary data to define bonding by that method
            4.2. Take best (defined by level of theory and electronic energy) task
            4.3. Convert TaskDoc to MoleculeBondingDoc
    """

    def __init__(
        self,
        tasks: Store,
        molecules: Store,
        bonds: Store,
        query: Optional[Dict] = None,
        methods: Optional[List] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        self.tasks = tasks
        self.molecules = molecules
        self.bonds = bonds
        self.query = query if query else dict()
        self.methods = methods if methods else BOND_METHODS
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks, molecules], targets=[bonds], **kwargs)
        # Uncomment in case of issue with mrun not connecting automatically to collections
        # for i in [self.tasks, self.molecules, self.bonds]:
        #     try:
        #         i.connect()
        #     except Exception as e:
        #         print("Could not connect,", e)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for bonds
        self.bonds.ensure_index("molecule_id")
        self.bonds.ensure_index("method")
        self.bonds.ensure_index("task_id")
        self.bonds.ensure_index("solvent")
        self.bonds.ensure_index("lot_solvent")
        self.bonds.ensure_index("property_id")
        self.bonds.ensure_index("last_updated")
        self.bonds.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.bonds.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into bonding documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the bonds Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Bonding builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.bonds.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into MoleculeBondingDocs

        Args:
            tasks List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new bonding docs
        """

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        bonding_docs = list()

        for mol in mols:
            correct_charge_spin = [
                e
                for e in mol.entries
                if e["charge"] == mol.charge
                and e["spin_multiplicity"] == mol.spin_multiplicity
            ]

            # Organize by solvent environment
            by_solvent = defaultdict(list)
            for entry in correct_charge_spin:
                by_solvent[entry["solvent"]].append(entry)

            for solvent, entries in by_solvent.items():
                sorted_entries = sorted(
                    entries,
                    key=lambda x: (
                        sum(evaluate_lot(x["level_of_theory"])),
                        x["energy"],
                    ),
                )

                for method in self.methods:
                    # For each method, grab entries that have the relevant data
                    if method == "OpenBabelNN + metal_edge_extender":
                        # This is sort of silly. Since, at the MoleculeDoc level,
                        # the structures have to be identical, bonding defined
                        # using heuristic methods like OpenBabel should always
                        # be identical.
                        # TODO: Decide if only one OpenBabelNN + m_e_e doc
                        # TODO: should be allowed.
                        relevant_entries = sorted_entries
                    else:
                        relevant_entries = [
                            e
                            for e in sorted_entries
                            if e.get(method) is not None
                            or e["output"].get(method) is not None
                        ]

                    if method == "nbo":
                        # Only allow NBO7 to be used. No earlier versions can be
                        # relied upon for bonding
                        relevant_entries = [
                            e
                            for e in relevant_entries
                            if e["orig"]["rem"].get("run_nbo6", False)
                            or e["orig"]["rem"].get("nbo_external", False)
                        ]

                    if len(relevant_entries) == 0:
                        continue

                    # Grab task document of best entry
                    best_entry = relevant_entries[0]
                    task = best_entry["task_id"]

                    tdoc = self.tasks.query_one(
                        {
                            "task_id": task,
                            "formula_alphabetical": formula,
                            "orig": {"$exists": True},
                        }
                    )

                    if tdoc is None:
                        try:
                            tdoc = self.tasks.query_one(
                                {
                                    "task_id": int(task),
                                    "formula_alphabetical": formula,
                                    "orig": {"$exists": True},
                                }
                            )
                        except ValueError:
                            tdoc = None

                    if tdoc is None:
                        continue

                    task_doc = TaskDocument(**tdoc)

                    if task_doc is None:
                        continue

                    doc = MoleculeBondingDoc.from_task(
                        task_doc,
                        molecule_id=mol.molecule_id,
                        preferred_methods=[method],
                        deprecated=False,
                    )
                    bonding_docs.append(doc)

        self.logger.debug(f"Produced {len(bonding_docs)} bonding docs for {formula}")

        return jsanitize([doc.dict() for doc in bonding_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new documents into the charges collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} bonding documents")
            self.bonds.remove_docs({self.bonds.key: {"$in": molecule_ids}})
            # Neither molecule_id nor method need to be unique, but the combination must be
            self.bonds.update(
                docs=docs,
                key=["molecule_id", "method", "solvent"],
            )
        else:
            self.logger.info("No items to update")



from setuptools import find_namespace_packages, setup

setup(
    name="emmet-builders",
    use_scm_version={"root": "..", "relative_to": __file__},
    setup_requires=["setuptools_scm"],
    description="Builders for the Emmet Library",
    author="The Materials Project",
    author_email="feedback@materialsproject.org",
    url="https://github.com/materialsproject/emmet",
    packages=find_namespace_packages(include=["emmet.*"]),
    install_requires=["emmet-core[all]", "maggma>=0.51.16", "matminer>=0.7.3"],
    extras_require={
        "test": [
            "pre-commit",
            "pytest",
            "pytest-cov",
            "pycodestyle",
            "pydocstyle",
            "flake8",
            "mypy",
            "mypy-extensions",
            "types-setuptools",
            "types-requests",
            "wincertstore",
        ],
        "docs": [
            "mkdocs",
            "mkdocs-material<8.3",
            "mkdocs-material-extensions",
            "mkdocs-minify-plugin",
            "mkdocstrings",
            "mkdocs-awesome-pages-plugin",
            "mkdocs-markdownextradata-plugin",
            "mkdocstrings[python]",
            "livereload",
            "jinja2",
        ],
    },
    python_requires=">=3.8",
    license="modified BSD",
    zip_safe=False,
)
